import warnings
from re import match
from time import sleep

import numpy as np
import pandas as pd
from future.utils import iteritems
from google.oauth2 import credentials as oauth2, service_account
from gspread.client import Client as ClientV4
from gspread.exceptions import APIError
from gspread.utils import a1_to_rowcol, rowcol_to_a1
from past.builtins import basestring

from gspread_pandas.exceptions import MissMatchException

ROW = START = 0
COL = END = 1
DEPRECATION_WARNINGS_ENABLED = True
_WARNINGS_ALREADY_ENABLED = False

# assuming no one will be 10 levels deep
auto_generated_index_names = ["level_{}".format(i) for i in range(10)] + ["index"]


def decode(strg):
    try:
        strg = strg.decode()
    except AttributeError:
        pass
    return strg


def parse_sheet_index(df, index):
    """Parse sheet index into df index."""
    if index and len(df.columns) >= index:
        df = df.set_index(df.columns[index - 1])
        # if column was MultiIndex, the name is a tuple;
        # choose last non-empty value in tuple
        # since that is more common
        if type(df.index.name) == tuple:
            df.index.name = [x for x in df.index.name if x][-1]
        # get rid of falsey index names
        df.index.name = df.index.name or None
    return df


def parse_df_col_names(df, include_index, index_size=1, flatten_sep=None):
    """Parse column names from a df into sheet headers."""
    headers = df.columns.tolist()

    # handle multi-index headers
    if len(headers) > 0 and type(headers[0]) == tuple:

        if isinstance(flatten_sep, basestring):
            headers = [
                [
                    # Remove blank elements and join using sep
                    flatten_sep.join([ele for ele in header if ele != ""])
                    for header in headers
                ]
            ]
        else:
            headers = [list(row) for row in zip(*headers)]

        # Pandas sets index name as top level col name when using reset_index;
        # move the index name to bottom level since that reads more natural
        if include_index:
            for i in range(index_size):
                top = headers[0]
                bottom = headers[-1]

                # Pandas sets the index's column name as "index" if it doesn't have a
                # name so we need to clean that up
                bottom[i] = top[i] if top[i] not in auto_generated_index_names else ""

                if len(headers) > 1:
                    top[i] = ""

    # handle regular columns
    else:
        headers = [headers]

    return headers


def parse_sheet_headers(vals, header_rows):
    """Parse headers from a sheet into df columns."""
    col_names = None
    if header_rows:
        headers = vals[:header_rows]
        if len(headers) > 0:
            if header_rows > 1:
                _fix_sheet_header_level(headers)
                col_names = pd.MultiIndex.from_arrays(headers)
            elif header_rows == 1:
                col_names = pd.Index(headers[0])

    return col_names


def _fix_sheet_header_level(header_names):
    for col_ix in range(len(header_names[0])):
        _shift_header_up(header_names, col_ix)

    return header_names


def _shift_header_up(
    header_names, col_index, row_index=0, shift_val=0, found_first=False
):
    """Recursively shift headers up so that top level is not empty."""
    rows = len(header_names)
    if row_index < rows:
        current_value = header_names[row_index][col_index]

        if current_value == "" and not found_first:
            shift_val += 1
        else:
            found_first = True

        shift_val = _shift_header_up(
            header_names, col_index, row_index + 1, shift_val, found_first
        )
        if shift_val <= row_index:
            header_names[row_index - shift_val][col_index] = current_value

        if row_index + shift_val >= rows:
            header_names[row_index][col_index] = ""
    return shift_val


def set_col_names(df, col_names):
    """Set the column names on the DataFrame and ensure the set even if there's no
    data."""
    if col_names is not None:
        if len(df.columns) == len(col_names):
            df.columns = col_names
        elif len(df) == 0:
            # if we have headers but no data, set column headers on empty DF
            df = df.reindex(columns=col_names)
        else:
            raise MissMatchException(
                "Column headers don't match number of data columns"
            )
    return df


def chunks(lst, chunk_size):
    """Chunk a list into specified chunk sizes."""
    for i in range(0, len(lst), chunk_size):
        yield lst[i : i + chunk_size]


def deprecate(message):
    """Display message about deprecation."""
    global DEPRECATION_WARNINGS_ENABLED, _WARNINGS_ALREADY_ENABLED
    # force enable DeprecationWarnings since most interactive shells have
    # them disabled
    if DEPRECATION_WARNINGS_ENABLED and not _WARNINGS_ALREADY_ENABLED:
        _WARNINGS_ALREADY_ENABLED = True
        warnings.filterwarnings(
            "default", ".*", category=DeprecationWarning, module="gspread_pandas"
        )
    # provide ability to disable them at runtime
    if _WARNINGS_ALREADY_ENABLED and not DEPRECATION_WARNINGS_ENABLED:
        warnings.filterwarnings(
            "ignore", ".*", category=DeprecationWarning, module="gspread_pandas"
        )

    warnings.warn(message, DeprecationWarning, stacklevel=2)


def create_filter_request(sheet_id, start, end):
    """Create v4 API request to create a filter for a given worksheet."""
    start = get_cell_as_tuple(start)
    end = get_cell_as_tuple(end)

    return {
        "setBasicFilter": {
            "filter": {
                "range": {
                    "sheetId": sheet_id,
                    "startRowIndex": start[ROW],
                    "endRowIndex": end[ROW],
                    "startColumnIndex": start[COL],
                    "endColumnIndex": end[COL],
                }
            }
        }
    }


def create_frozen_request(sheet_id, rows=None, cols=None):
    """Create v4 API request to freeze rows and/or columns for a given worksheet."""
    grid_properties = {}

    if rows is not None and rows >= 0:
        grid_properties["frozen_row_count"] = rows

    if cols is not None and cols >= 0:
        grid_properties["frozen_column_count"] = cols

    changed_props = grid_properties.keys()

    return {
        "update_sheet_properties": {
            "properties": {"sheet_id": sheet_id, "grid_properties": grid_properties},
            "fields": "grid_properties({0})".format(", ".join(changed_props)),
        }
    }


def fillna(df, fill_value=""):
    """
    Replace null values with `fill_value`.

    Also replaces in categorical columns.
    """
    for col in df.dtypes[df.dtypes == "category"].index:
        if fill_value not in df[col].cat.categories:
            df[col].cat.add_categories([fill_value], inplace=True)
    return df.fillna(fill_value)


def get_cell_as_tuple(cell):
    """Take cell in either format, validate, and return as tuple."""
    if type(cell) == tuple:
        if (
            len(cell) != 2
            or not np.issubdtype(type(cell[ROW]), np.integer)
            or not np.issubdtype(type(cell[COL]), np.integer)
        ):
            raise TypeError("{0} is not a valid cell tuple".format(cell))
        return cell
    elif isinstance(cell, basestring):
        if not match("[a-zA-Z]+[0-9]+", cell):
            raise TypeError("{0} is not a valid address".format(cell))
        return a1_to_rowcol(cell)
    else:
        raise TypeError("{0} is not a valid format".format(cell))


def get_range(start, end):
    """Transform start and end to cell range like A1:B5."""
    start_int = get_cell_as_tuple(start)
    end_int = get_cell_as_tuple(end)

    return "{0}:{1}".format(rowcol_to_a1(*start_int), rowcol_to_a1(*end_int))


def create_merge_cells_request(sheet_id, start, end, merge_type="MERGE_ALL"):
    """Create v4 API request to merge rows and/or columns for a given worksheet."""
    start = get_cell_as_tuple(start)
    end = get_cell_as_tuple(end)

    return {
        "mergeCells": {
            "range": {
                "sheetId": sheet_id,
                "startRowIndex": start[ROW] - 1,
                "endRowIndex": end[ROW],
                "startColumnIndex": start[COL] - 1,
                "endColumnIndex": end[COL],
            },
            "mergeType": merge_type,
        }
    }


def create_unmerge_cells_request(sheet_id, start, end):
    """Create v4 API request to unmerge rows and/or columns for a given worksheet."""
    start = get_cell_as_tuple(start)
    end = get_cell_as_tuple(end)

    return {
        "unmergeCells": {
            "range": {
                "sheetId": sheet_id,
                "startRowIndex": start[ROW] - 1,
                "endRowIndex": end[ROW],
                "startColumnIndex": start[COL] - 1,
                "endColumnIndex": end[COL],
            }
        }
    }


def monkey_patch_request(client, retry_delay=10):
    """Monkey patch gspread's Client.request to auto-retry with a delay when you get a
    100 seconds RESOURCE_EXCHAUSTED error."""

    def request(*args, **kwargs):
        try:
            return ClientV4.request(client, *args, **kwargs)
        except APIError as e:
            error = str(e)
            # Only retry on 100 seconds quota breaches
            if "RESOURCE_EXHAUSTED" in error and "100" in error:
                sleep(retry_delay)
                return request(*args, **kwargs)
            else:
                error = e

        if "error" in locals():
            raise error

    client.request = request


def create_merge_headers_request(sheet_id, headers, start, index_size):
    """Create v4 API request to merge labels for a given worksheet."""
    request = []
    start = get_cell_as_tuple(start)

    if isinstance(headers, pd.MultiIndex):
        merge_cells = get_col_merge_ranges(headers)
        request.append(
            [
                create_merge_cells_request(
                    sheet_id,
                    (start[ROW] + row_ix, col_rng[START] + start[COL] + index_size),
                    (start[ROW] + row_ix, col_rng[END] + start[COL] + index_size),
                )
                for row_ix, row in enumerate(merge_cells)
                for col_rng in row
            ]
        )

    return request


def get_col_merge_ranges(index):
    """
    Get list of ranges to be merged for each level of columns.

    For each level, same values will only be merged if they share the
    same label for the level above.
    """
    labels = index.codes if hasattr(index, "codes") else index.labels
    # Dummy range indicating the full size, this is removed at the end
    ranges = [[(0, len(labels[0]))]]

    for ix, index_level in enumerate(labels[:]):
        index_ranges = []
        for rng in ranges[ix]:
            index_ranges.extend(
                get_contiguous_ranges(index_level, rng[START], rng[END])
            )
        ranges.append(index_ranges)

    ranges.pop(0)
    return ranges


def get_contiguous_ranges(lst, lst_start, lst_end):
    """
    Get list of tuples, each indicating the range of contiguous equal values in the lst
    between lst_start and lst_end. Everything is 0 indexed.

    For example, get_contiguous_ranges([0, 0, 0, 1, 1], 1, 4) = [(1, 2), (3, 4)]
    [(the 2nd and 3rd items are both 0), (the 4th and 5th items are both 1)]
    """
    prev_val = None
    index_ranges = []
    lst_section = lst[lst_start : lst_end + 1]
    rng_start = lst_section[START]

    for ix, val in enumerate(lst_section):
        # If there's more than 1 value in the range and there's a change in val
        if rng_start < ix - 1 and prev_val != val:
            # Save the rng until the previous val
            index_ranges.append((lst_start + rng_start, lst_start + ix - 1))
        # If this is the last val and the val is still the same
        elif ix == len(lst_section) - 1 and prev_val == val:
            # Save the rng including the last val
            index_ranges.append((lst_start + rng_start, lst_start + ix))

        if prev_val != val:
            rng_start = ix

        prev_val = val
    return index_ranges


def convert_credentials(credentials):
    """Convert oauth2client credentials to google-auth."""
    cls = credentials.__class__.__name__

    if cls == "ServiceAccountCredentials":
        return _convert_service_account(credentials)
    elif cls == "OAuth2Credentials":
        return _convert_oauth(credentials)

    raise TypeError(
        "Credentials need to be from either oauth2client or from google-auth."
    )


def _convert_oauth(credentials):
    return oauth2.Credentials(
        credentials.access_token,
        credentials.refresh_token,
        credentials.id_token,
        credentials.token_uri,
        credentials.client_id,
        credentials.client_secret,
        credentials.scopes,
    )


def _convert_service_account(credentials):
    data = credentials.serialization_data
    data["token_uri"] = credentials.token_uri
    scopes = credentials._scopes.split() or [
        "https://www.googleapis.com/auth/drive",
        "https://spreadsheets.google.com/feeds",
    ]

    return service_account.Credentials.from_service_account_info(data, scopes=scopes)


def parse_permission(perm):
    """Convert the string permission into a dict to unpack for insert_permission."""
    perm_dict = {}
    perm = perm.split("|")
    for part in perm:
        if "@" in part:
            perm_dict["value"] = part
            perm_dict["perm_type"] = "user"
        elif "." in part:
            perm_dict["value"] = part
            perm_dict["perm_type"] = "domain"
        elif "anyone" == part:
            perm_dict["perm_type"] = "anyone"
        elif part in ["grp", "group"]:
            perm_dict["perm_type"] = "group"
        elif part in ["owner", "writer", "reader"]:
            perm_dict["role"] = part
        elif part in ["no", "false"]:
            perm_dict["notify"] = False
        elif part == "link":
            perm_dict["with_link"] = True
        perm_dict["role"] = perm_dict.get("role", "reader")
    return perm_dict


def remove_keys(dct, keys=[]):
    """Remove keys from a dict."""
    return {key: val for key, val in iteritems(dct) if key not in keys}


def remove_keys_from_list(lst, keys=[]):
    """Remove keys from a list of dicts."""
    return [remove_keys(ele, keys) for ele in lst]


def add_paths(root, dirs):
    """
    Recursively build a `path` property to each dir.

    Pass in the root dir and a list of all available dirs.
    """
    # TODO: handle scenario with folders having more than one parent
    children = [dr for dr in dirs if root.get("id", None) in dr.get("parents", [])]
    path = root.get("path", "")

    for child in children:
        child["path"] = path + "/" + child["name"]
        add_paths(child, dirs)


def folders_to_create(search_path, dirs, base_path=""):
    """
    Recursively traverse through folder paths looking for the longest existing subpath.

    Return the dir info of the longest subpath and the directories that
    need to be created.
    """
    # Allow user to pass in a string, but use a list in the recursion
    if isinstance(search_path, list):
        parts = search_path
    else:
        parts = search_path.strip("/").split("/")

    parent = [dr for dr in dirs if dr.get("path", "") == base_path]
    if len(parent) == 0:
        parent = {"id": "root"}
    else:
        parent = parent.pop()

    # Stop if we ran out of parts to create
    if len(parts) == 0:
        return parent, []

    base_path += "/" + parts[0]

    dirs = [dr for dr in dirs if dr.get("path", "").startswith(base_path)]

    # If there's base_path matches, then keep looking for a longer path
    if len(dirs) > 0:
        return folders_to_create(parts[1:], dirs, base_path)
    else:
        return parent, parts


def get_ranges(sheet_name, cols):
    """Get a list of ranges for the given worksheet and columns."""
    ranges = []
    for col in cols:
        start_label = rowcol_to_a1(1, col)
        range_label = "%s!%s:%s" % (sheet_name, start_label, start_label[:-1])
        ranges.append(range_label)
    return ranges


def is_int(val):
    return isinstance(val, (int, np.integer))


def is_indexes(lst):
    """Is this a list of indexes (all ints)"""
    return all([is_int(val) for val in lst])


def find_col_indexes(cols, col_names, col_offset=1):
    """Given a column name Index, find the numeric indeces of the columns in the
    spreadsheet."""
    col_locs = []

    for col in cols:
        loc = col_names.get_loc(col)
        if is_int(loc):
            col_locs.append(loc)
        elif isinstance(loc, slice):
            col_locs += list(range(len(col_names))[loc])
        elif isinstance(loc, np.ndarray):
            col_locs += [ix for ix in range(len(loc)) if loc[ix]]
    # add 1 because we want the index based on spreadsheet, not python
    return [ele + col_offset for ele in set(col_locs)]



from __future__ import print_function

from builtins import range, str
from re import match

import numpy as np
import pandas as pd
from gspread.exceptions import (
    APIError,
    NoValidUrlKeyFound,
    SpreadsheetNotFound,
    WorksheetNotFound,
)
from gspread.models import Worksheet
from gspread.utils import fill_gaps, rightpad
from past.builtins import basestring

from gspread_pandas.client import Client
from gspread_pandas.conf import default_scope
from gspread_pandas.exceptions import (
    GspreadPandasException,
    MissMatchException,
    NoWorksheetException,
)
from gspread_pandas.util import (
    COL,
    ROW,
    chunks,
    create_filter_request,
    create_frozen_request,
    create_merge_cells_request,
    create_merge_headers_request,
    create_unmerge_cells_request,
    deprecate,
    fillna,
    find_col_indexes,
    get_cell_as_tuple,
    get_range,
    get_ranges,
    is_indexes,
    parse_df_col_names,
    parse_permission,
    parse_sheet_headers,
    parse_sheet_index,
    set_col_names,
)

__all__ = ["Spread"]


class Spread:
    """
    Simple wrapper for gspread to interact with Pandas. It holds an instance of an
    'open' spreadsheet, an 'open' worksheet, and a list of available worksheets.

    Each user will be associated with specific OAuth credentials. The authenticated user
    will need the appropriate permissions to the Spreadsheet in order to interact with
    it.

    Parameters
    ----------
    spread : str
        name, url, or id of the spreadsheet; must have read access by
        the authenticated user,
        see :meth:`open_spread <gspread_pandas.spread.Spread.open_spread>`
    sheet : str,int
        optional, name or index of Worksheet,
        see :meth:`open_sheet <gspread_pandas.spread.Spread.open_sheet>`
        (default None)
    config : dict
        optional, if you want to provide an alternate configuration,
        see :meth:`get_config <gspread_pandas.conf.get_config>` (default None)
    create_sheet : bool
        whether to create the worksheet if it doesn't exist,
        it wil use the ``spread`` value as the sheet title (default False)
    create_spread : bool
        whether to create the spreadsheet if it doesn't exist,
        it wil use the ``spread`` value as the sheet title (default False)
    scope : list
        optional, if you'd like to provide your own scope
        (default default_scope)
    user : str
        string indicating the key to a users credentials,
        which will be stored in a file (by default they will be stored in
        ``~/.config/gspread_pandas/creds/<user>`` but can be modified with
        ``creds_dir`` property in config). If using a Service Account, this
        will be ignored. (default "default")
    creds : google.auth.credentials.Credentials
        optional, pass credentials if you have those already (default None)
    client : Client
        optionall, if you've already instanciated a Client, you can just pass
        that and it'll be used instead (default None)
    permissions : list
        a list of strings. See
        :meth:`add_permissions <gspread_pandas.spread.Spread.add_permissions>`
        for the expected format
    """

    #: `(gspread.models.Spreadsheet)` - Currently open Spreadsheet
    spread = None

    #: `(gspread.models.Worksheet)` - Currently open Worksheet
    sheet = None

    #: `(Client)` - Instance of gspread_pandas
    #: :class:`Client <gspread_pandas.client.Client>`
    client = None

    # chunk range request: https://github.com/burnash/gspread/issues/375
    _max_range_chunk_size = 1000000

    # `(dict)` - Spreadsheet metadata
    _spread_metadata = None

    def __init__(
        self,
        spread,
        sheet=0,
        config=None,
        create_spread=False,
        create_sheet=False,
        scope=default_scope,
        user="default",
        creds=None,
        client=None,
        permissions=None,
    ):
        if isinstance(client, Client):
            self.client = client
        else:
            self.client = Client(user, config, scope, creds)

        self.open(spread, sheet, create_sheet, create_spread)

        if permissions:
            self.add_permissions(permissions)

    def __repr__(self):
        base = "<gspread_pandas.spread.Spread - '{}'>"
        meta = []
        if self.email:
            meta.append("User: '{}'".format(self.email))
        if self.spread:
            meta.append("Spread: '{}'".format(self.spread.title))
        if self.sheet:
            meta.append("Sheet: '{}'".format(self.sheet.title))
        return base.format(", ".join(meta))

    def __iter__(self):
        for sheet in self.sheets:
            yield sheet

    @property
    def email(self):
        """`(str)` - E-mail for the currently authenticated user"""
        return self.client.email

    @property
    def url(self):
        """`(str)` - Url for this spreadsheet"""
        return "https://docs.google.com/spreadsheets/d/{}".format(self.spread.id)

    @property
    def sheets(self):
        """`(list)` - List of available Worksheets"""
        return self.spread.worksheets()

    def refresh_spread_metadata(self):
        """Refresh spreadsheet metadata."""
        self._spread_metadata = self.spread.fetch_sheet_metadata()

    @property
    def _sheet_metadata(self):
        """`(dict)` - Metadata for currently open worksheet"""
        if self.sheet:
            ix = self._find_sheet(self.sheet.title)[0]
            return self._spread_metadata["sheets"][ix]

    def open(self, spread, sheet=None, create_sheet=False, create_spread=False):
        """
        Open a spreadsheet, and optionally a worksheet. See.

        :meth:`open_spread <gspread_pandas.Spread.open_spread>` and
        :meth:`open_sheet <gspread_pandas.Spread.open_sheet>`.

        Parameters
        ----------
        spread : str
            name, url, or id of Spreadsheet
        sheet : str,int
            name or index of Worksheet (default None)
        create_sheet : bool
            whether to create the worksheet if it doesn't exist,
            it wil use the ``spread`` value as the sheet title (default False)
        create_spread : bool
            whether to create the spreadsheet if it doesn't exist,
            it wil use the ``spread`` value as the sheet title (default False)

        Returns
        -------
        None
        """
        self.open_spread(spread, create_spread)

        if sheet is not None:
            self.open_sheet(sheet, create_sheet)

    def open_spread(self, spread, create=False):
        """
        Open a spreadsheet. Authorized user must already have read access.

        Parameters
        ----------
        spread : str
            name, url, or id of Spreadsheet
        create : bool
            whether to create the spreadsheet if it doesn't exist,
            it wil use the ``spread`` value as the sheet title (default False)

        Returns
        -------
        None
        """
        id_regex = "[a-zA-Z0-9-_]{44}"
        url_path = "docs.google.com/spreadsheet"

        if match(id_regex, spread):
            open_func = self.client.open_by_key
        elif url_path in spread:
            open_func = self.client.open_by_url
        else:
            open_func = self.client.open

        try:
            self.spread = open_func(spread)
            self.refresh_spread_metadata()
        except (SpreadsheetNotFound, NoValidUrlKeyFound, APIError) as error:
            if create:
                try:
                    self.spread = self.client.create(spread)
                    self.refresh_spread_metadata()
                except Exception as e:
                    msg = "Couldn't create spreadsheet.\n" + str(e)
                    new_error = GspreadPandasException(msg)
            elif isinstance(error, SpreadsheetNotFound) or "NOT_FOUND" in str(error):
                new_error = SpreadsheetNotFound("Spreadsheet not found")
            else:
                new_error = error

        # Raise new exception outside of except block for a python2/3 way to avoid
        # "During handling of the above exception, another exception occurred"
        if "new_error" in locals() and isinstance(new_error, Exception):
            raise new_error

    def open_sheet(self, sheet, create=False):
        """
        Open a worksheet. Optionally, if the sheet doesn't exist then create it first
        (only when ``sheet`` is a str).

        Parameters
        ----------
        sheet : str,int,Worksheet
            name, index, or Worksheet object
        create : bool
            whether to create the sheet if it doesn't exist,
            see :meth:`create_sheet <gspread_pandas.Spread.create_sheet>`
            (default False)

        Returns
        -------
        None
        """
        self.sheet = None
        if isinstance(sheet, int):
            if sheet >= len(self.sheets) or sheet < -1 * len(self.sheets):
                raise WorksheetNotFound("Invalid sheet index {}".format(sheet))
            self.sheet = self.sheets[sheet]
        else:
            self.sheet = self.find_sheet(sheet)

        if not self.sheet:
            if create:
                self.create_sheet(sheet)
            else:
                raise WorksheetNotFound("Worksheet not found")

    def create_sheet(self, name, rows=1, cols=1):
        """
        Create a new worksheet with the given number of rows and cols.

        Automatically opens that sheet after it's created.

        Parameters
        ----------
        name : str
            name of new Worksheet
        rows : int
            number of rows (default 1)
        cols : int
            number of columns (default 1)

        Returns
        -------
        None
        """
        self.spread.add_worksheet(name, rows, cols)
        self.refresh_spread_metadata()
        self.open_sheet(name)

    def _get_columns(self, cols, value_render_option="FORMATTED_VALUE"):
        """
        Returns a list of all values in `cols`.

        Empty cells in this list will be rendered as :const:`None`.

        Parameters
        ----------
        cols : list of ints
            Column numbers.
        value_render_option : str
            Determines how values should be rendered in the the output. Possible
            values are "FORMATTED_VALUE", "FORMULA", and "UNFORMATTED_VALUE"
            (Default value = "FORMATTED_VALUE")

        Returns
        -------
        """
        ranges = get_ranges(self.sheet.title, cols)
        data = self.spread.values_batch_get(
            ranges,
            params={
                "valueRenderOption": value_render_option,
                "majorDimension": "COLUMNS",
            },
        )

        try:
            return fill_gaps(
                [col.get("values", [[]])[0] for col in data["valueRanges"]]
            )
        except KeyError:
            return []

    def _fix_value_render(
        self, df, first_data_row, col_names, cols, value_render_option
    ):
        """Replace values for columns that need a different value render option."""
        if not is_indexes(cols):
            cols = find_col_indexes(cols, col_names)

        for ix, col in enumerate(self._get_columns(cols, value_render_option)):
            df.iloc[:, cols[ix] - 1] = rightpad(col[first_data_row:], len(df))

    def sheet_to_df(
        self,
        index=1,
        header_rows=1,
        start_row=1,
        unformatted_columns=None,
        formula_columns=None,
        sheet=None,
    ):
        """
        Pull a worksheet into a DataFrame.

        Parameters
        ----------
        index : int
            col number of index column, 0 or None for no index (default 1)
        header_rows : int
            number of rows that represent headers (default 1)
        start_row : int
            row number for first row of headers or data (default 1)
        unformatted_columns : list
            column numbers or names for columns you'd like to pull in as
            unformatted values (defaul [])
        formula_columns : list
            column numbers or names for columns you'd like to pull in as
            actual formulas (defaul [])
        sheet : str,int
            optional, if you want to open a different sheet first,
            see :meth:`open_sheet <gspread_pandas.spread.Spread.open_sheet>`
            (default None)

        Returns
        -------
        DataFrame
            DataFrame with the data from the Worksheet
        """
        self._ensure_sheet(sheet)

        vals = self.sheet.get_all_values()
        vals = self._fix_merge_values(vals)[start_row - 1 :]

        col_names = parse_sheet_headers(vals, header_rows)

        # remove rows where everything is null, then replace nulls with ''
        df = (
            pd.DataFrame(vals[header_rows or 0 :])
            .replace("", np.nan)
            .dropna(how="all")
            .fillna("")
        )

        # replace values with a different value render option before we set the
        # index in set_col_names
        if unformatted_columns:
            self._fix_value_render(
                df,
                header_rows + start_row - 1,
                col_names,
                unformatted_columns,
                "UNFORMATTED_VALUE",
            )

        if formula_columns:
            self._fix_value_render(
                df, header_rows + start_row - 1, col_names, formula_columns, "FORMULA"
            )

        df = set_col_names(df, col_names)

        return parse_sheet_index(df, index)

    def get_sheet_dims(self, sheet=None):
        """
        Get the dimensions of the currently open Worksheet.

        Parameters
        ----------
        sheet : str,int,Worksheet
            optional, if you want to open a different sheet first,
            see :meth:`open_sheet <gspread_pandas.spread.Spread.open_sheet>`
            (default None)

        Returns
        -------
        tuple
            a tuple containing (num_rows,num_cols)
        """
        self._ensure_sheet(sheet)
        return (self.sheet.row_count, self.sheet.col_count) if self.sheet else None

    def _get_update_chunks(self, start, end, vals):
        start = get_cell_as_tuple(start)
        end = get_cell_as_tuple(end)

        num_cols = end[COL] - start[COL] + 1
        num_rows = end[ROW] - start[ROW] + 1
        num_cells = num_cols * num_rows

        if num_cells != len(vals):
            raise MissMatchException("Number of values needs to match number of cells")

        chunk_rows = self._max_range_chunk_size // num_cols
        chunk_size = chunk_rows * num_cols

        end_cell = (start[ROW] - 1, 0)

        for val_chunks in chunks(vals, int(chunk_size)):
            start_cell = (end_cell[ROW] + 1, start[COL])
            end_cell = (
                min(start_cell[ROW] + chunk_rows - 1, start[ROW] + num_rows - 1),
                end[COL],
            )
            yield start_cell, end_cell, val_chunks

    def update_cells(self, start, end, vals, sheet=None, raw_columns=None):
        """
        Update the values in a given range. The values should be listed in order from
        left to right across rows.

        Parameters
        ----------
        start : tuple,str
            tuple indicating (row, col) or string like 'A1'
        end : tuple,str
            tuple indicating (row, col) or string like 'Z20'
        vals : list
            array of values to populate
        sheet : str,int,Worksheet
            optional, if you want to open a different sheet first,
            see :meth:`open_sheet <gspread_pandas.spread.Spread.open_sheet>`
            (default None)
        raw_columns : list, int
            optional, list of column numbers in the google sheet that should be
            interpreted as "RAW" input

        Returns
        -------
        None
        """
        self._ensure_sheet(sheet)

        for start_cell, end_cell, val_chunks in self._get_update_chunks(
            start, end, vals
        ):
            rng = get_range(start_cell, end_cell)

            cells = self.sheet.range(rng)

            if len(val_chunks) != len(cells):
                raise MissMatchException(
                    "Number of chunked values doesn't match number of cells"
                )

            for val, cell in zip(val_chunks, cells):
                cell.value = val

            if raw_columns:
                assert isinstance(
                    raw_columns, list
                ), "raw_columns must be a list of ints"
                raw_cells = [i for i in cells if i.col in raw_columns]
                self.sheet.update_cells(raw_cells, "RAW")
            else:
                raw_cells = []

            user_cells = [i for i in cells if i not in raw_cells]
            if user_cells:
                self.sheet.update_cells(user_cells, "USER_ENTERED")

    def _ensure_sheet(self, sheet):
        if sheet is not None:
            self.open_sheet(sheet, create=True)

        if not self.sheet:
            raise NoWorksheetException("No open worksheet")

    def _find_sheet(self, sheet):
        """
        Find a worksheet and return with index.

        Parameters
        ----------
        sheet : str,Worksheet
            Name or worksheet to find


        Returns
        -------
        tuple
            Tuple like (index, worksheet)
        """
        for ix, worksheet in enumerate(self.sheets):
            if (
                isinstance(sheet, basestring)
                and sheet.lower() == worksheet.title.lower()
            ):
                return ix, worksheet
            if isinstance(sheet, Worksheet) and sheet.id == worksheet.id:
                return ix, worksheet
        return None, None

    def find_sheet(self, sheet):
        """
        Find a given worksheet by title or by object comparison.

        Parameters
        ----------
        sheet : str,Worksheet
            name of Worksheet or Worksheet object

        Returns
        -------
        Worksheet
            the Worksheet by the given name or None if not found
        """
        return self._find_sheet(sheet)[1]

    def clear_sheet(self, rows=1, cols=1, sheet=None):
        """
        Reset open worksheet to a blank sheet with given dimensions.

        Parameters
        ----------
        rows : int
            number of rows (default 1)
        cols : int
            number of columns (default 1)
        sheet : str,int,Worksheet
            optional; name, index, or Worksheet,
            see :meth:`open_sheet <gspread_pandas.spread.Spread.open_sheet>`
            (default None)

        Returns
        -------
        None
        """
        self._ensure_sheet(sheet)

        # TODO: if my merge request goes through, use sheet.frozen_*_count
        frozen_rows = self._sheet_metadata["properties"]["gridProperties"].get(
            "frozenRowCount", 0
        )
        frozen_cols = self._sheet_metadata["properties"]["gridProperties"].get(
            "frozenColCount", 0
        )

        row_resize = max(rows, frozen_rows + 1)
        col_resize = max(cols, frozen_cols + 1)

        self.sheet.resize(row_resize, col_resize)

        self.update_cells(
            start=(1, 1),
            end=(row_resize, col_resize),
            vals=["" for i in range(0, row_resize * col_resize)],
        )

    def delete_sheet(self, sheet):
        """
        Delete a worksheet by title. Returns whether the sheet was deleted or not. If
        current sheet is deleted, the ``sheet`` property will be set to None.

        Parameters
        ----------
        sheet : str,Worksheet
            name or Worksheet

        Returns
        -------
        bool
            True if deleted successfully, else False
        """
        is_current = False

        s = self.find_sheet(sheet)

        if s == self.sheet:
            is_current = True

        if s:
            try:
                self.spread.del_worksheet(s)
                if is_current:
                    self.sheet = None
                return True
            except Exception:
                pass

        self.refresh_spread_metadata()

        return False

    def df_to_sheet(
        self,
        df,
        index=True,
        headers=True,
        start=(1, 1),
        replace=False,
        sheet=None,
        raw_column_names=None,
        raw_columns=None,
        freeze_index=False,
        freeze_headers=False,
        fill_value="",
        add_filter=False,
        merge_headers=False,
        flatten_headers_sep=None,
    ):
        """
        Save a DataFrame into a worksheet.

        Parameters
        ----------
        df : DataFrame
            the DataFrame to save
        index : bool
            whether to include the index in worksheet (default True)
        headers : bool
            whether to include the headers in the worksheet (default True)
        start : tuple,str
            tuple indicating (row, col) or string like 'A1' for top left
            cell (default (1,1))
        replace : bool
            whether to remove everything in the sheet first (default False)
        sheet : str,int,Worksheet
            optional, if you want to open or create a different sheet
            before saving,
            see :meth:`open_sheet <gspread_pandas.spread.Spread.open_sheet>`
            (default None)
        raw_column_names : list, str
            (DEPRECATED use raw_collumns instead) optional, list of columns
            from your dataframe that you want interpreted as RAW input in
            google sheets.
        raw_columns : list, str
            optional, list of columns from your dataframe that you want
            interpreted as RAW input in google sheets. This can be column
            names or column numbers.
        freeze_index : bool
            whether to freeze the index columns (default False)
        freeze_headers : bool
            whether to freeze the header rows (default False)
        fill_value : str
            value to fill nulls with (default '')
        add_filter : bool
            whether to add a filter to the uploaded sheet (default False)
        merge_headers : bool
            whether to merge cells in the header that have the same value
            (default False)
        flatten_headers_sep : str
            if you want to flatten your multi-headers to a single row,
            you can pass the string that you'd like to use to concatenate
            the levels, for example, ': ' (default None)

        Returns
        -------
        None
        """
        self._ensure_sheet(sheet)

        header = df.columns
        index_size = df.index.nlevels if index else 0
        header_size = df.columns.nlevels

        if index:
            df = df.reset_index()

        df = fillna(df, fill_value)
        df_list = df.values.tolist()

        if headers:
            header_rows = parse_df_col_names(df, index, index_size, flatten_headers_sep)
            df_list = header_rows + df_list

        start = get_cell_as_tuple(start)

        sheet_rows, sheet_cols = self.get_sheet_dims()
        req_rows = len(df_list) + (start[ROW] - 1)
        req_cols = len(df_list[0]) + (start[COL] - 1) or 1

        end = (req_rows, req_cols)

        if replace:
            # this takes care of resizing
            self.clear_sheet(req_rows, req_cols)
        else:
            # make sure sheet is large enough
            self.sheet.resize(max(sheet_rows, req_rows), max(sheet_cols, req_cols))

        if raw_column_names:
            deprecate("raw_column_names is deprecated, please use raw_columns instead.")
            raw_columns = find_col_indexes(
                raw_column_names, header, start[COL] + index_size
            )
        elif raw_columns:
            if is_indexes(raw_columns):
                offset = index_size + start[COL] - 1
                raw_columns = [ix + offset for ix in raw_columns]
            else:
                raw_columns = find_col_indexes(
                    raw_columns, header, start[COL] + index_size
                )

        self.update_cells(
            start=start,
            end=end,
            vals=[str(val) for row in df_list for val in row],
            raw_columns=raw_columns,
        )

        self.freeze(
            None if not freeze_headers else header_size + start[ROW] - 1,
            None if not freeze_index else index_size + start[COL] - 1,
        )

        if add_filter:
            self.add_filter(
                (header_size + start[ROW] - 2, start[COL] - 1), (req_rows, req_cols)
            )

        if merge_headers:
            self.spread.batch_update(
                {
                    "requests": create_merge_headers_request(
                        self.sheet.id, header, start, index_size
                    )
                }
            )

        self.refresh_spread_metadata()

    def _fix_merge_values(self, vals):
        """
        Assign the top-left value to all cells in a merged range.

        Parameters
        ----------
        vals : list
            Values returned by
            :meth:`get_all_values() <gspread.models.Sheet.get_all_values()>_`


        Returns
        -------
        list
            Fixed values
        """
        for merge in self._sheet_metadata.get("merges", []):
            start_row, end_row = merge["startRowIndex"], merge["endRowIndex"]
            start_col, end_col = (merge["startColumnIndex"], merge["endColumnIndex"])

            # ignore merge cells outside the data range
            if start_row < len(vals) and start_col < len(vals[0]):
                orig_val = vals[start_row][start_col]
                for row in vals[start_row:end_row]:
                    row[start_col:end_col] = [
                        orig_val for i in range(start_col, end_col)
                    ]

        return vals

    def freeze(self, rows=None, cols=None, sheet=None):
        """
        Freeze rows and/or columns for the open worksheet.

        Parameters
        ----------
        rows : int
            number of rows to freeze, use 0 to 'unfreeze' (default None)
        cols : int
            number of columns to freeze, use 0 to 'unfreeze' (default None)
        sheet : str,int,Worksheet
            optional, if you want to open or create a
            different sheet before freezing,
            see :meth:`open_sheet <gspread_pandas.spread.Spread.open_sheet>`
            (default None)

        Returns
        -------
        None
        """
        self._ensure_sheet(sheet)

        if rows is None and cols is None:
            return

        self.spread.batch_update(
            {"requests": create_frozen_request(self.sheet.id, rows, cols)}
        )

        self.refresh_spread_metadata()

    def add_filter(self, start=None, end=None, sheet=None):
        """
        Add filters to data in the open worksheet.

        Parameters
        ----------
        start : tuple,str
            Tuple indicating (row, col) or string like 'A1' (default 'A1')
        end : tuple, str
            Tuple indicating (row, col) or string like 'A1'
            (default last cell in sheet)
        sheet : str,int,Worksheet
            optional, if you want to open or create a
            different sheet before adding the filter,
            see :meth:`open_sheet <gspread_pandas.spread.Spread.open_sheet>`
            (default None)

        Returns
        -------
        None
        """
        self._ensure_sheet(sheet)

        dims = self.get_sheet_dims()

        self.spread.batch_update(
            {
                "requests": create_filter_request(
                    self.sheet.id, start or (0, 0), end or dims
                )
            }
        )

        self.refresh_spread_metadata()

    def merge_cells(self, start, end, merge_type="MERGE_ALL", sheet=None):
        """
        Merge cells between the start and end cells. Use merge_type if you want to
        change the behavior of the merge.

        Parameters
        ----------
        start : tuple,str
            Tuple indicating (row, col) or string like 'A1'
        end : tuple, str
            Tuple indicating (row, col) or string like 'A1'
        merge_type : str
            One of MERGE_ALL, MERGE_ROWS, or MERGE_COLUMNS (default "MERGE_ALL")
        sheet : str,int,Worksheet
            optional, if you want to open or create a
            different sheet before adding the filter,
            see :meth:`open_sheet <gspread_pandas.spread.Spread.open_sheet>`
            (default None)

        Returns
        -------
        None
        """
        self._ensure_sheet(sheet)

        self.spread.batch_update(
            {"requests": create_merge_cells_request(self.sheet.id, start, end)}
        )

        self.refresh_spread_metadata()

    def unmerge_cells(self, start="A1", end=None, sheet=None):
        """
        Unmerge all cells between the start and end cells. Use defaults to unmerge all
        cells in the sheet.

        Parameters
        ----------
        start : tuple,str
            Tuple indicating (row, col) or string like 'A1' (default A1)
        end : tuple,str
            Tuple indicating (row, col) or string like 'A1' (default last cell in sheet)
        sheet : str,int,Worksheet
            optional, if you want to open or create a
            different sheet before adding the filter,
            see :meth:`open_sheet <gspread_pandas.spread.Spread.open_sheet>`
            (default None)

        Returns
        -------
        None
        """
        self._ensure_sheet(sheet)

        if end is None:
            end = self.get_sheet_dims()

        self.spread.batch_update(
            {"requests": create_unmerge_cells_request(self.sheet.id, start, end)}
        )

        self.refresh_spread_metadata()

    def add_permission(self, permission):
        """
        Add a permission to the current spreadsheet.

        The format should be:
        ``<id>|(<group>)|(<role>)|(<notify>)|(<require_link>)`` where:

        <id> - email address of group or individual, domain, or 'anyone'
        <group> - optional, if the id is a group e-mail, this needs to be 'group' or
            'grp'
        <role> - optional, one of 'owner', 'writer', or 'reader'. If ommited, 'reader'
            will be used
        <notify> - optional, if you don't want to notify the user, pass 'no' or 'false'
        <require_link> - optional, if you want to require the user to have the link,
            pass 'link'

        For example, to allow anyone with a link in the group admins@example.com to
        write when they have a link, but without sending a notification to the group:
        ``admins@example.com|grp|owner|false|link``

        Or if you want to give user@example.com reader permissions without a
        notification:
        ``user@example.com|no``

        Or to give anyone read access:
        ``anyone``

        Parameters
        ----------
        permissions : string
            A strings meeting the above mentioned format.


        Returns
        -------
        None
        """
        perm = parse_permission(permission)
        self.client.insert_permission(self.spread.id, perm.pop("value", None), **perm)

    def add_permissions(self, permissions):
        """
        Add permissions to the current spreadsheet. See.

        :meth:`add_permission <gspread_pandas.spread.Spread.add_permission>` for format.


        Parameters
        ----------
        permissions : list
            A list of strings meeting the above mentioned format.


        Returns
        -------
        None
        """
        for perm in permissions:
            self.add_permission(perm)

    def list_permissions(self):
        """
        List all permissions for this Spreadsheet.

        Returns
        -------
        list
            a list of dicts indicating the permissions on this spreadsheet
        """
        return self.client.list_permissions(self.spread.id)

    def move(self, path="/", create=True):
        """
        Move the current spreadsheet to the specified path in your Google drive. If the
        file is not currently in you drive, it will be added.

        Parameters
        ----------
        path : str
            folder path (Default value = "/")
        create : bool
            if true, create folders as needed (Default value = True)

        Returns
        -------
        """
        self.client.move_file(self.spread.id, path, create)



import json
import sys
from os import environ, name

from future.utils import reraise
from google.oauth2.credentials import Credentials as OAuthCredentials
from google.oauth2.service_account import Credentials as SACredentials
from google_auth_oauthlib.flow import InstalledAppFlow
from past.builtins import basestring

from gspread_pandas.exceptions import ConfigException
from gspread_pandas.util import decode

try:
    from pathlib import Path
except ImportError:
    from pathlib2 import Path


__all__ = ["default_scope", "get_config", "get_creds"]
if name == "nt":
    _default_dir = Path(environ.get("APPDATA")) / "gspread_pandas"
else:
    _default_dir = (
        Path(environ.get("XDG_CONFIG_HOME", Path(environ.get("HOME", "")) / ".config"))
        / "gspread_pandas"
    )
_default_file = "google_secret.json"

default_scope = [
    "openid",
    "https://www.googleapis.com/auth/drive",
    "https://www.googleapis.com/auth/userinfo.email",
    "https://www.googleapis.com/auth/spreadsheets",
]

CONFIG_DIR_ENV_VAR = "GSPREAD_PANDAS_CONFIG_DIR"


def get_config_dir():
    """
    Get the config directory.

    It will first look in the environment variable
    GSPREAD_PANDAS_CONFIG_DIR, but if it's not set it'll use
    ~/.config/gspread_pandas
    """
    return Path(environ.get(CONFIG_DIR_ENV_VAR, _default_dir))


def ensure_path(full_path):
    """
    Create path if it doesn't exist.

    Parameters
    ----------
    full_path : str
        Path to create if needed

    Returns
    -------
    None
    """
    full_path = Path(full_path)
    if not full_path.exists():
        full_path.mkdir(parents=True, exist_ok=True)


def get_config(conf_dir=None, file_name=_default_file):
    """
    Get config for Google client. Looks in ~/.config/gspread_pandas/google_secret.json
    by default but you can override it with conf_dir and file_name. The creds_dir value
    will be set to conf_dir/creds and the directory will be created if it doesn't exist;
    if you'd like to override that you can do so by changing the 'creds_dir' value in
    the dict returned by this function.

    Download json from https://console.developers.google.com/apis/credentials

    Parameters
    ----------
    conf_dir : str
        Full path to config dir (Default value = get_config_dir())
    file_name : str
        (Default value = "google_secret.json")

    Returns
    -------
    dict
        Dict with necessary contents of google_secret.json
    """
    conf_dir = Path(conf_dir).expanduser() if conf_dir else get_config_dir()
    cfg_file = conf_dir / file_name

    if not cfg_file.exists():
        raise IOError(
            "No Google client config found.\n"
            "Please download json from "
            "https://console.developers.google.com/apis/credentials and "
            "save as {}".format(cfg_file)
        )

    with cfg_file.open() as fp:
        cfg = json.load(fp)

    return cfg


def get_creds(
    user="default", config=None, scope=default_scope, creds_dir=None, save=True
):
    """
    Get google google.auth.credentials.Credentials for the given user. If the user
    doesn't have previous creds, they will go through the OAuth flow to get new
    credentials which will be saved for later use. Credentials will be saved in
    config['creds_dir'], if this value is not set, then they will be stored in a folder
    named ``creds`` in the default config dir (either ~/.config/gspread_pandas or.

    $GSPREAD_PANDAS_CONFIG_DIR)

    Alternatively, it will get credentials from a service account.

    Parameters
    ----------
    user : str
        Unique key indicating user's credentials. This is not necessary when using
        a ServiceAccount and will be ignored (Default value = "default")
    config : dict
        Optional, dict with "client_id", "client_secret", and "redirect_uris" keys for
        OAuth or "type", "client_email", "private_key", "private_key_id", and
        "client_id" for a Service Account. If None is passed, it will call
        :meth:`get_config() <get_config>` (Default value = None)
    creds_dir : str, Path
        Optional, directory to load and store creds from/in. If None, it will use the
        ``creds`` subdirectory in the default config location. (Default value = None)
    scope : list
        Optional, scope to use for Google Auth (Default value = default_scope)

    Returns
    -------
    google.auth.credentials.Credentials
        Google credentials that can be used with gspread
    """
    config = config or get_config()
    try:
        if "private_key_id" in config:
            return SACredentials.from_service_account_info(config, scopes=scope)

        if not isinstance(user, basestring):
            raise ConfigException(
                "Need to provide a user key as a string if not using a service account"
            )

        if creds_dir is None:
            creds_dir = get_config_dir() / "creds"

        creds_file = Path(creds_dir) / user

        if creds_file.exists():
            # need to convert Path to string for python 2.7
            return OAuthCredentials.from_authorized_user_file(str(creds_file))

        flow = InstalledAppFlow.from_client_config(
            config, scope, redirect_uri="urn:ietf:wg:oauth:2.0:oob"
        )
        creds = flow.run_console()

        if save:
            creds_data = {
                "refresh_token": creds.refresh_token,
                "token_uri": creds.token_uri,
                "client_id": creds.client_id,
                "client_secret": creds.client_secret,
                "scopes": creds.scopes,
            }

            ensure_path(creds_dir)
            creds_file.write_text(decode(json.dumps(creds_data)))

        return creds
    except Exception:
        exc_info = sys.exc_info()

    if "exc_info" in locals():
        reraise(ConfigException, *exc_info[1:])



from gspread.models import Spreadsheet
from gspread.urls import SPREADSHEETS_API_V4_BASE_URL

from ._version import __version__, __version_info__
from .client import Client
from .spread import Spread

__all__ = ["Spread", "Client", "__version__", "__version_info__"]
SPREADSHEET_VALUES_BATCH_URL = SPREADSHEETS_API_V4_BASE_URL + "/%s/values:batchGet"


def values_batch_get(self, ranges, params=None):
    """
    Lower-level method that directly calls `spreadsheets.values.batchGet.

    <https://develop
    ers.google.com/sheets/api/reference/rest/v4/spreadsheets.values/batchGet>`_.

    Parameters
    ----------
    rantes : list of strings
        List of ranges in the `A1 notation
    params : dict
        (optional) Query parameters

    Returns
    -------
    dict
        Response body
    """
    if params is None:
        params = {}

    params.update(ranges=ranges)

    url = SPREADSHEET_VALUES_BATCH_URL % (self.id)
    r = self.client.request("get", url, params=params)
    return r.json()


Spreadsheet.values_batch_get = values_batch_get



class GspreadPandasException(Exception):
    pass


class ConfigException(GspreadPandasException):
    pass


class NoWorksheetException(GspreadPandasException):
    pass


class MissMatchException(GspreadPandasException):
    pass



from __future__ import print_function

from builtins import super

import requests
from future.utils import reraise
from google.auth.credentials import Credentials
from google.auth.transport.requests import AuthorizedSession
from gspread.client import Client as ClientV4
from gspread.exceptions import APIError, SpreadsheetNotFound
from gspread.models import Spreadsheet
from gspread.utils import finditem
from past.builtins import basestring

from gspread_pandas.conf import default_scope, get_creds
from gspread_pandas.util import (
    add_paths,
    convert_credentials,
    folders_to_create,
    monkey_patch_request,
    remove_keys_from_list,
)

__all__ = ["Client"]


class Client(ClientV4):
    """
    The gspread_pandas :class:`Client` extends :class:`Client <gspread.client.Client>`
    and authenticates using credentials stored in ``gspread_pandas`` config.

    This class also adds a few convenience methods to explore the user's google drive
    for spreadsheets.

    Parameters
    ----------
    user : str
        optional, string indicating the key to a users credentials,
        which will be stored in a file (by default they will be stored in
        ``~/.config/gspread_pandas/creds/<user>`` but can be modified with
        ``creds_dir`` property in config). If using a Service Account, this
        will be ignored. (default "default")
    config : dict
        optional, if you want to provide an alternate configuration,
        see :meth:`get_config <gspread_pandas.conf.get_config>`
        (default None)
    scope : list
        optional, if you'd like to provide your own scope
        (default default_scope)
    creds : google.auth.credentials.Credentials
        optional, pass credentials if you have those already (default None)
    session : google.auth.transport.requests.AuthorizedSession
        optional, pass a google.auth.transport.requests.AuthorizedSession or a
        requests.Session and creds (default None)
    load_dirs : bool
        optional, whether you want to load directories and paths on instanciation.
        if you refresh directories later or perform an action that requires them,
        they will be loaded at that time. For speed, this is disabled by default
        (default False)
    """

    _email = None
    _root = None
    _dirs = None
    _load_dirs = False

    def __init__(
        self,
        user="default",
        config=None,
        scope=default_scope,
        creds=None,
        session=None,
        load_dirs=False,
    ):
        #: `(list)` - Feeds included for the OAuth2 scope
        self.scope = scope

        if isinstance(session, requests.Session):
            credentials = getattr(session, "credentials", creds)
            if not credentials:
                raise TypeError(
                    "If you provide a session, you must also provide credentials"
                )
        else:
            if isinstance(creds, Credentials):
                credentials = creds
            elif creds is not None and "oauth2client" in creds.__module__:
                credentials = convert_credentials(creds)
            elif isinstance(user, basestring):
                credentials = get_creds(user, config, self.scope)
            else:
                raise TypeError(
                    "Need to provide user as a string or credentials as "
                    "google.auth.credentials.Credentials"
                )
            session = AuthorizedSession(credentials)
        super().__init__(credentials, session)

        monkey_patch_request(self)

        self._root = self._drive_request(file_id="root", params={"fields": "name,id"})

        if load_dirs:
            self.refresh_directories()

    @property
    def root(self):
        """`(dict)` - the info for the top level Drive directory for current user"""
        return self._root

    def _get_dirs(self, strip_parents=True):
        """
        Helper function to fetch directories if they haven't been yet.

        It will strip the parents by default for the `directories`
        property
        """
        if not self._load_dirs:
            self.refresh_directories()

        if strip_parents:
            # this will make a copy, if we intend to modify the values
            # internally, pass strip_parents = False
            return remove_keys_from_list(self._dirs, ["parents"])
        else:
            return self._dirs

    directories = property(
        _get_dirs,
        doc=(
            "`(list)` - list of dicts for all avaliable "
            "directories for the current user"
        ),
    )

    @property
    def email(self):
        """`(str)` - E-mail for the currently authenticated user"""
        if not self._email:
            try:
                self._email = self.request(
                    "get", "https://www.googleapis.com/userinfo/v2/me"
                ).json()["email"]
            except Exception:
                print(
                    """
                    Couldn't retrieve email. Delete credentials and authenticate again
                    """
                )

        return self._email

    def refresh_directories(self):
        """Refresh list of directories for the current user."""
        self._load_dirs = True
        q = "mimeType='application/vnd.google-apps.folder'"
        self._dirs = self._query_drive(q)
        add_paths(self._root, self._dirs)

    def login(self):
        """Override login since AuthorizedSession now takes care of automatically
        refreshing tokens when needed."""

    def _query_drive(self, q):
        files = []
        page_token = ""
        params = {"q": q, "pageSize": 1000, "fields": "files(name,id,parents)"}

        while page_token is not None:
            if page_token:
                params["pageToken"] = page_token

            res = self._drive_request("get", params=params)
            files.extend(res.get("files", []))
            page_token = res.get("nextPageToken", None)

        return files

    def _drive_request(
        self, method="get", file_id=None, params=None, data=None, headers=None
    ):
        url = "https://www.googleapis.com/drive/v3/files"
        if file_id:
            url += "/{}".format(file_id)
        try:
            res = self.request(method, url, params=params, json=data)
            if res.text:
                return res.json()
        except APIError as e:
            if "scopes" in e.response.text:
                print(
                    "Your credentials don't have Drive API access, ignoring "
                    "drive specific functionality (Note this includes searching "
                    "spreadsheets by name)"
                )
                return {}
            else:
                reraise(e)

    def open(self, title):
        """
        Opens a spreadsheet.

        :param title: A title of a spreadsheet.
        :type title: str

        :returns: a :class:`~gspread.models.Spreadsheet` instance.

        If there's more than one spreadsheet with same title the first one
        will be opened.

        :raises gspread.SpreadsheetNotFound: if no spreadsheet with
                                             specified `title` is found.

        >>> c = gspread.authorize(credentials)
        >>> c.open('My fancy spreadsheet')
        """
        try:
            properties = finditem(
                lambda x: x["name"] == title, self.list_spreadsheet_files(title)
            )

            # Drive uses different terminology
            properties["title"] = properties["name"]

            return Spreadsheet(self, properties)
        except StopIteration:
            raise SpreadsheetNotFound

    def list_spreadsheet_files(self, title=None):
        """
        Return all spreadsheets that the user has access to.

        Parameters
        ----------
        title : str
            name of the spreadsheet, if none is passed it'll return every file
            (default None)

        Returns
        -------
        list
            List of spreadsheets. Each spreadsheet is a dict with the following keys:
            id, kind, mimeType, and name.
        """
        q = "mimeType='application/vnd.google-apps.spreadsheet'"
        if title:
            q += ' and name = "{}"'.format(title)
        return self._list_spreadsheet_files(q)

    def list_spreadsheet_files_in_folder(self, folder_id):
        """
        Return all spreadsheets that the user has access to in a sepcific folder.

        Parameters
        ----------
        folder_id : str
            ID of a folder, see :meth:`find_folders <find_folders>`

        Returns
        -------
        list
            List of spreadsheets. Each spreadsheet is a dict with the following keys:
            id, kind, mimeType, and name.
        """
        q = (
            "mimeType='application/vnd.google-apps.spreadsheet'"
            " and '{}' in parents".format(folder_id)
        )

        return self._list_spreadsheet_files(q)

    def _list_spreadsheet_files(self, q):
        """Helper function to actually run a query, add paths if needed, and remove
        unwanted keys from results."""
        files = self._query_drive(q)

        if self._load_dirs:
            self._add_path_to_files(files)

        return remove_keys_from_list(files, ["parents"])

    def _add_path_to_files(self, files):
        """Add path to files by looking up the parent dir and its path."""
        for fil3 in files:
            try:
                # if a file is in multiple directories then it'll
                # have multiple parents. However, this adds complexity
                # so we'll just choose the first one to build the path
                parent = next(
                    directory
                    for directory in self._get_dirs(False) + [self._root]
                    if directory["id"] in fil3.get("parents", {})
                )
                fil3["path"] = parent.get("path", "/")
            except StopIteration:
                # Files that are visible to a ServiceAccount but not
                # in the root will not have the 'parents' property
                fil3["path"] = None

    def find_folders(self, folder_name_query=""):
        """
        Return all folders that the user has access to containing ``folder_name_query``
        in the name.

        Parameters
        ----------
        folder_name_query : str
            Case insensitive string to search in folder name. If empty,
            it will return all folders.

        Returns
        -------
        list
            List of folders. Each folder is a dict with the following keys:
            id, kind, mimeType, and name.
        """
        return [
            folder
            for folder in self.directories
            if folder_name_query.lower() in folder["name"].lower()
        ]

    def find_spreadsheet_files_in_folders(self, folder_name_query):
        """
        Return all spreadsheets that the user has access to in all the folders that
        contain ``folder_name_query`` in the name. Returns as a dict with each key being
        the folder name and the value being a list of spreadsheet files.

        Parameters
        ----------
        folder_name_query : str
            Case insensitive string to search in folder name

        Returns
        -------
        dict
            Spreadsheets in each folder. Each entry is a dict with the folder name as
            the key and a list of spreadsheets as the value. Each spreadsheet is a dict
            with the following keys: id, kind, mimeType, and name.
        """

        return {
            res["name"]: self.list_spreadsheet_files_in_folder(res["id"])
            for res in self.find_folders(folder_name_query)
        }

    def create_folder(self, path, parents=True):
        """
        Create a new folder in your Google drive.

        Parameters
        ----------
        path : str
            folder path
        parents : bool
            if True, create parent folders as needed (Default value = True)

        Returns
        -------
        dict
            information for the created directory
        """
        parent, to_create = folders_to_create(path, self._get_dirs(False))

        if len(to_create) > 1 and parents is not True:
            raise Exception(
                "If you want to create nested directories pass parents=True"
            )

        for directory in to_create:
            parent = self._drive_request(
                "post",
                params={"fields": "name,id,parents"},
                data={
                    "mimeType": "application/vnd.google-apps.folder",
                    "name": directory,
                    "parents": [parent["id"]],
                },
                headers={"Content-Type": "application/json"},
            )

        self.refresh_directories()
        return parent

    def move_file(self, file_id, path, create=False):
        """
        Move a file to the given path.

        Parameters
        ----------
        file_id : str
            file id
        path : str
            folder path
        create : bool
            whether to create any missing folders (Default value = False)

        Returns
        -------
        """
        if path == "/":
            folder_id = "root"
        else:
            parent, missing = folders_to_create(path, self._get_dirs(False))
            if missing:
                if not create:
                    raise Exception("Folder does not exist")

                parent = self.create_folder(path)
            folder_id = parent["id"]

        old_parents = self._drive_request(
            "get", file_id, params={"fields": "parents"}
        ).get("parents", [])

        params = {"addParents": folder_id, "removeParents": ",".join(old_parents)}
        self._drive_request("patch", file_id, params)
